import os
import os.path
import importlib


def get_name():
    mod = importlib.import_module("config.project")
    if hasattr(mod, "name"):
        return getattr(mod, "name")
    return os.path.basename(os.getcwd())



"""
This simply extracts the description of a package.

References:
- https://docs.python.org/3/library/importlib.metadata.html
"""

from importlib.metadata import metadata


def get_package_description(name: str) -> str:
    s = metadata(name)
    return s["Summary"]



"""
This module build python virtual envrionments
"""


import os
import shutil
from typing import List, Generator, Tuple

from pydmt.api.builder import Builder, Node, SourceFile, TargetFolder
from pydmt.utils.filesystem import files_under_folder
from pydmt.utils.digest import sha1_file
from pydmt.utils.subprocess import check_call
from pydmt.utils.python import collect_reqs, get_install_args

SOURCE_FILE = "config/python.py"
TARGET_FOLDER = ".venv/default"


class BuilderVenvFull(Builder):
    """
    This is a review of how to build a python virtual environment:
    # create the virtualenv
    virtualenv [folder]
    # activate it
    source [folder]/bin/activate
    # install package
    pip install -r requirements.txt
    """
    def __init__(self):
        # pylint: disable=useless-super-delegation
        super().__init__()

    def get_sources(self) -> List[Node]:
        file_list = [SourceFile(SOURCE_FILE)]
        return file_list

    def get_targets(self) -> List[Node]:
        return [TargetFolder(TARGET_FOLDER)]

    def build(self) -> None:
        if os.path.isdir(TARGET_FOLDER):
            shutil.rmtree(TARGET_FOLDER)
        args = [
            "virtualenv",
            TARGET_FOLDER,
        ]
        check_call(args)
        args = [
            "venv-run",
            "--venv",
            ".venv/default",
            "--",
        ]
        args.extend(get_install_args())
        packs = collect_reqs()
        if packs:
            args.extend(packs)
            check_call(args)

    def yield_results(self) -> Generator[Tuple[str, str], None, None]:
        for x in files_under_folder(TARGET_FOLDER):
            yield sha1_file(x), x



"""
This module build python virtual envrionments
"""


from pydmt.api.one_source_one_target import OneSourceOneTarget
from pydmt.utils.subprocess import check_call
from pydmt.utils.filesystem import mkdir_touch
from pydmt.utils.python import collect_reqs, collect_bootstrap_reqs, get_install_args


class BuilderReqs(OneSourceOneTarget):
    """
    This is a review of how to build a python virtual environment:
    # create the virtualenv
    virtualenv [folder]
    # activate it
    source [folder]/bin/activate
    # install packages from file
    pip install -r requirements.txt
    # OR
    python -m pip install -r requirements.txt
    # OR if you want to install a list of packages:
    pip install -r [list of packages]
    # OR
    python -m pip install [list of packages]
    """
    def build(self) -> None:
        args = get_install_args()
        packs = collect_bootstrap_reqs()
        if packs:
            args.extend(packs)
            check_call(args)
        args = get_install_args()
        packs = collect_reqs()
        if packs:
            args.extend(packs)
            check_call(args)
        mkdir_touch(self.target)



import sys
import os
import os.path
from typing import List, Dict, Union, Generator, Tuple

import mako
import mako.exceptions
import mako.lookup
import mako.template

from pydmt.api.builder import Builder, Node, SourceFile, TargetFile, SourceFolder
from pydmt.utils.filesystem import makedirs_for_file
from pydmt.utils.digest import sha1_file

FOLDER_CONFIG = "config"
FOLDER_SNIPPLETS = "snipplets"


class BuilderMako(Builder):
    def __init__(self,
                 source: str,
                 target: str,
                 data: Union[Dict[str, object], None],
                 config_files: List[str],
                 snipplet_files: List[str],
                 ):
        super().__init__()
        self.source = source
        self.target = target
        self.data = data
        self.config_files: List[str] = config_files
        self.snipplet_files: List[str] = snipplet_files
        self.sources = [SourceFile(self.source)]
        if os.path.isdir(FOLDER_CONFIG):
            self.sources.append(SourceFolder(FOLDER_CONFIG))
        if os.path.isdir(FOLDER_SNIPPLETS):
            self.sources.append(SourceFolder(FOLDER_SNIPPLETS))
        self.targets = [TargetFile(self.target)]

    def get_sources(self) -> List[Node]:
        return self.sources

    def get_targets(self) -> List[Node]:
        return self.targets

    def build(self):
        lookup = mako.lookup.TemplateLookup(
            directories=['.'],
        )
        template = mako.template.Template(
            filename=self.source,
            lookup=lookup,
        )
        makedirs_for_file(self.target)
        if self.data is None:
            output = template.render()
        else:
            output = template.render(**self.data)
        with open(self.target, "w") as file_handle:
            file_handle.write(output)

    def yield_results(self) -> Generator[Tuple[str, str], None, None]:
        yield sha1_file(self.target), self.target


def print_full_exception():
    print("printing full exception")
    traceback = mako.exceptions.RichTraceback()
    for (filename, line_number, function_name, line) in traceback.traceback:
        print(f"File {filename}, line {line_number}, in {function_name}")
        print(line)
    print(f"{traceback.error.__class__.__name__}: {traceback.error}")


def print_exception(e, input_file):
    found = False
    traceback = mako.exceptions.RichTraceback()
    for (filename, line_number, function_name, line) in traceback.traceback:
        if filename == input_file:
            print(f"{sys.argv[0]}: error {e} in {filename}, line {line_number} function {function_name}")
            print(f"{line}")
            found = True
    if not found:
        for (filename, line_number, function_name, line) in traceback.traceback:
            print(f"File {filename}, line {line_number}, in {function_name}")
            print(line)
        print(f"{traceback.error.__class__.__name__}: {traceback.error}")



"""
This is a module that will install OS packages for you.
"""


from typing import List
import os

from pydmt.utils.filesystem import unlink_files, mkdir_touch
from pydmt.configs import ConfigSudo
from pydmt.utils.subprocess import check_call

from pydmt.api.one_source_one_target import OneSourceOneTarget


class BuilderApt(OneSourceOneTarget):
    def __init__(self, source: str, target: str, packages: List[str]):
        super().__init__(source, target)
        self.packages = packages

    def build(self) -> None:
        unlink_files(self.target)
        os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
        args = []
        if ConfigSudo.sudo:
            args.append("sudo")
        args.extend([
            'apt-get',
            '-q=2',
            '--yes',
            'update',
        ])
        check_call(args)
        args = []
        if ConfigSudo.sudo:
            args.append("sudo")
        args.extend([
            'apt-get',
            '-q=2',
            '--yes',
            'install',
        ])
        args.extend(self.packages)
        check_call(args)
        mkdir_touch(self.target)



import sys
import os
import pickle
import pathlib
import shutil
import logging

import pylogconf.core
from pytconf import register_endpoint, register_main, config_arg_parse_and_launch, get_free_args

from pydmt.configs import ConfigSudo, ConfigFlow, ConfigOutput, ConfigTarget, ConfigLogging
from pydmt.core.pydmt import PyDMT
from pydmt.static import APP_NAME, VERSION_STR, DESCRIPTION, LOGGER_NAME
from pydmt.utils.subprocess import check_call

from pydmt.features.sphinx import FeatureSphinx
from pydmt.features.mako import FeatureMako
from pydmt.features.yaml import FeatureYaml
from pydmt.features.apt import FeatureApt
from pydmt.features.npm import FeatureNpm
from pydmt.features.venv import FeatureVenv
from pydmt.features.reqs import FeatureReqs
from pydmt.features.make import FeatureMake


def add_all_features(p):
    f = FeatureApt()
    f.setup(p)
    f = FeatureNpm()
    f.setup(p)
    if ConfigTarget.dev:
        f = FeatureVenv()
        f.setup(p)
        f = FeatureMako()
        f.setup(p)
        f = FeatureSphinx()
        f.setup(p)
    else:
        f = FeatureReqs()
        f.setup(p)
    f = FeatureMake()
    f.setup(p)
    f = FeatureYaml()
    f.setup(p)


def add_to_path():
    """
    This adds to PYTHONPATH various paths we need
    If you disable this then templates would not be able to find things like 'config/python.py'
    """
    add_import_of_cwd = True
    add_import_of_home = True
    add_import_of_shared = True
    if add_import_of_shared:
        folder = "/etc/pydmt"
        if folder not in sys.path:
            sys.path.insert(0, folder)
    if add_import_of_home:
        folder = os.path.join(str(pathlib.Path.home()), ".config/pydmt")
        if folder not in sys.path:
            sys.path.insert(0, folder)
    if add_import_of_cwd:
        if "" not in sys.path:
            sys.path.insert(0, "")


@register_endpoint(
    description="Build the project",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigTarget,
        ConfigLogging,
    ],
)
def build():
    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    add_all_features(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Build python virtual environment",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigLogging,
    ],
)
def build_venv():
    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    f = FeatureVenv()
    f.setup(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Install python prerequisites",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigLogging,
    ],
)
def build_reqs():
    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    f = FeatureReqs()
    f.setup(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Build tools",
    configs=[
        ConfigSudo,
        ConfigFlow,
        ConfigOutput,
        ConfigLogging,
    ],
)
def build_tools():
    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    f = FeatureApt()
    f.setup(p)
    f = FeatureNpm()
    f.setup(p)

    stats = p.build_all()
    sys.exit(stats.get_os_error_code())


@register_endpoint(
    description="Clean all generated files",
    configs=[
        ConfigLogging,
    ],
)
def clean() -> None:
    add_to_path()

    pylogconf.core.setup()
    p = PyDMT()

    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    add_all_features(p)

    shutil.rmtree(".pydmt")
    p.clean_all()


@register_endpoint(
    description="Clean all files and pydmt cache",
    configs=[
        ConfigLogging,
    ],
)
def clean_hard() -> None:
    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(ConfigLogging.loglevel)

    shutil.rmtree(".pydmt")
    check_call(["git", "clean", "-qffxd"])


@register_endpoint(
    description="print a pickle file for debug purposes",
    configs=[
    ],
    allow_free_args=True,
)
def print_pickle() -> None:
    for filename in get_free_args():
        with open(filename, "rb") as file_handle:
            print(f"{filename}")
            for x, y in pickle.load(file_handle).items():
                print(x, y)


@register_main(
    app_name=APP_NAME,
    version=VERSION_STR,
    main_description=DESCRIPTION,
)
def main():
    pylogconf.core.setup()
    config_arg_parse_and_launch()


if __name__ == '__main__':
    main()



import hashlib
from typing import List

from pydmt.utils.filesystem import files_under_folders


BLOCK_SIZE = 65536
ALGORITHM_NAME = "sha1"


class Digester:
    def __init__(self):
        self.hash_object = hashlib.new(ALGORITHM_NAME)

    def add_file(self, filename: str) -> None:
        with open(filename, 'rb') as file_handle:
            buf = file_handle.read(BLOCK_SIZE)
            while len(buf) > 0:
                self.hash_object.update(buf)
                buf = file_handle.read(BLOCK_SIZE)

    def add_files(self, filenames: List[str]) -> None:
        for filename in filenames:
            self.add_file(filename)

    def add_folders(self, folders: List[str]) -> None:
        for filename in files_under_folders(folders=folders):
            self.add_file(filename)

    def add_files_folders(self, files: List[str], folders: List[str]) -> None:
        f = files
        f.extend(files_under_folders(folders=folders))
        self.add_files(f)

    def get_hexdigest(self):
        return self.hash_object.hexdigest()



import logging

from pytconf import Config, ParamCreator


class ConfigSudo(Config):
    """
    Parameters to control whether we use 'sudo' or not
    """
    sudo = ParamCreator.create_bool(
        help_string="use sudo?",
        default=True,
    )


class ConfigFlow(Config):
    """
    Parameters to configure the flow of pydmt
    """
    stop_after_error = ParamCreator.create_bool(
        help_string="Should pydmt stop after first error?",
        default=True,
    )


class ConfigTarget(Config):
    """
    Parameters to configure what is the target of the build
    """
    dev = ParamCreator.create_bool(
        help_string="Is the target to build a dev environment?",
        default=True,
    )


class ConfigOutput(Config):
    """
    Parameters to configure the output of pydmt
    """
    verbose = ParamCreator.create_bool(
        help_string="Should output be verbose?",
        default=True,
    )
    print_not = ParamCreator.create_bool(
        help_string="print out what we are not doing",
        default=False,
    )


class ConfigLogging(Config):
    """
    Parameters to control logging
    """
    loglevel = ParamCreator.create_choice(
        choice_list=[
            logging.getLevelName(logging.NOTSET),
            logging.getLevelName(logging.DEBUG),
            logging.getLevelName(logging.INFO),
            logging.getLevelName(logging.WARNING),
            logging.getLevelName(logging.WARN),
            logging.getLevelName(logging.ERROR),
            logging.getLevelName(logging.FATAL),
            logging.getLevelName(logging.CRITICAL),
        ],
        help_string="What log level to use?",
        default=logging.getLevelName(logging.INFO),
    )


class ConfigSubprocess(Config):
    """
    Parameters to configure how we run subprocess
    """
    print_command = ParamCreator.create_bool(
        help_string="print out commands",
        default=True,
    )
    quiet = ParamCreator.create_bool(
        help_string="Suppress output?",
        default=False,
    )



import os

from pydmt.api.feature import Feature
from pydmt.builders.venv_full import BuilderVenvFull
from pydmt.core.pydmt import PyDMT


class FeatureVenv(Feature):
    def setup(self, pydmt: PyDMT) -> None:
        if not os.path.isfile("config/python.py"):
            return
        pydmt.add_builder(BuilderVenvFull())



import importlib


def get_workflows_platforms():
    try:
        mod = importlib.import_module("config.github")
        if hasattr(mod, "workflows_platforms"):
            return getattr(mod, "workflows_platforms")
    except ModuleNotFoundError:
        pass
    mod = importlib.import_module("default.github")
    if hasattr(mod, "workflows_platforms"):
        return getattr(mod, "workflows_platforms")
    raise ValueError("cannot find worklows platformas")



import os
import glob
import pprint
from typing import List

from pydmt.configs import ConfigTarget


def hlp_source_under(folder):
    """
    this function finds all the python packages under a folder and
    write the 'packages' and 'package_dir' entries for a python setup.py
    script
    """
    # walk the folder and find the __init__.py entries for packages.
    packages = []
    package_dir = {}
    for root, _, files in os.walk(folder):
        for file in files:
            if file != '__init__.py':
                continue
            full = os.path.dirname(os.path.join(root, file))
            relative = os.path.relpath(full, folder)
            packages.append(relative)
            package_dir[relative] = full
    # we use pprint because we want the order to always remain the same
    return f"packages={sorted(packages)},\npackage_dir={pprint.pformat(package_dir)}"


def hlp_files_under(destination_folder, pat):
    file_list = [x for x in glob.glob(pat) if os.path.isfile(x)]
    return f"('{destination_folder}', {file_list})"


def make_hlp_project_keywords(d):
    def hlp_project_keywords():
        return f"{d.project_keywords.split()}"

    return hlp_project_keywords


def make_hlp_project_platforms(d):
    def hlp_project_platforms():
        return f"{d.project_platforms.split()}"

    return hlp_project_platforms


def make_hlp_project_classifiers(d):
    def hlp_project_classifiers():
        classifiers = d.project_classifiers.split('\n')
        classifiers = [x.strip()[1:-1] for x in classifiers]
        return f"{classifiers}"

    return hlp_project_classifiers


def make_hlp_wrap(level):
    def hlp_wrap(t):
        return t.replace('\n', '\n' + '\t' * level)

    return hlp_wrap


def collect_reqs() -> List[str]:
    collect = []
    # pylint: disable=import-outside-toplevel,no-name-in-module,import-error,no-member
    import config.python
    if hasattr(config.python, "dev_requires") and ConfigTarget.dev:
        collect.extend(config.python.dev_requires)
    if hasattr(config.python, "test_requires"):
        collect.extend(config.python.test_requires)
    if hasattr(config.python, "install_requires"):
        collect.extend(config.python.install_requires)
    if hasattr(config.python, "config_requires"):
        collect.extend(config.python.config_requires)
    if hasattr(config.python, "setup_requires"):
        collect.extend(config.python.setup_requires)
    if hasattr(config.python, "make_requires"):
        collect.extend(config.python.make_requires)
    return collect


def collect_bootstrap_reqs() -> List[str]:
    # pylint: disable=import-outside-toplevel,no-name-in-module,import-error,no-member
    import config.bootstrap
    return config.bootstrap.requires


def get_install_args():
    return [
        "python",
        "-m",
        "pip",
        "install",
    ]
